@@ -29,6 +29,7 @@
 #include "gkr-keyring-item.h"
 
 #include "common/gkr-buffer.h"
+#include "common/gkr-location.h"
 
 #include "daemon/gnome-keyring-daemon.h"
 
@@ -60,7 +61,8 @@ enum {
 
 enum {
     PROP_0,
-    PROP_NAME
+    PROP_NAME,
+    PROP_LOCATION
 };
 
 static guint signals[LAST_SIGNAL] = { 0 };
@@ -293,12 +295,12 @@ generate_acl_data (GkrBuffer *buffer,
 	GList *l;
 	GnomeKeyringAccessControl *ac;
 	
-	gnome_keyring_proto_add_uint32 (buffer, g_list_length (acl));
+	gkr_buffer_add_uint32 (buffer, g_list_length (acl));
 
 	for (l = acl; l != NULL; l = l->next) {
 		ac = l->data;
 		
-		gnome_keyring_proto_add_uint32 (buffer, ac->types_allowed);
+		gkr_buffer_add_uint32 (buffer, ac->types_allowed);
 		if (!gnome_keyring_proto_add_utf8_string (buffer, ac->application->display_name)) {
 			return FALSE;
 		}
@@ -309,7 +311,7 @@ generate_acl_data (GkrBuffer *buffer,
 		if (!gnome_keyring_proto_add_utf8_string (buffer, NULL)) {
 			return FALSE;
 		}
-		gnome_keyring_proto_add_uint32 (buffer, 0);
+		gkr_buffer_add_uint32 (buffer, 0);
 	}
 	
 	
@@ -342,7 +344,7 @@ generate_encrypted_data (GkrBuffer *buffer, GkrKeyring *keyring)
 			return FALSE;
 		}
 		for (i = 0; i < 4; i++) {
-			gnome_keyring_proto_add_uint32 (buffer, 0);
+			gkr_buffer_add_uint32 (buffer, 0);
 		}
 
 		if (!gnome_keyring_proto_add_attribute_list (buffer, item->attributes)) {
@@ -388,23 +390,23 @@ generate_file (GkrBuffer *buffer, GkrKeyring *keyring)
 	if (keyring->lock_on_idle) {
 		flags |= 1;
 	}
-	gnome_keyring_proto_add_uint32 (buffer, flags);
-	gnome_keyring_proto_add_uint32 (buffer, keyring->lock_timeout);
-	gnome_keyring_proto_add_uint32 (buffer, keyring->hash_iterations);
+	gkr_buffer_add_uint32 (buffer, flags);
+	gkr_buffer_add_uint32 (buffer, keyring->lock_timeout);
+	gkr_buffer_add_uint32 (buffer, keyring->hash_iterations);
 	gkr_buffer_append (buffer, (guchar*)keyring->salt, 8);
 
 	/* Reserved: */
 	for (i = 0; i < 4; i++) {
-		gnome_keyring_proto_add_uint32 (buffer, 0);
+		gkr_buffer_add_uint32 (buffer, 0);
 	}
 
 	/* Hashed items: */
-	gnome_keyring_proto_add_uint32 (buffer, g_list_length (keyring->items));
+	gkr_buffer_add_uint32 (buffer, g_list_length (keyring->items));
 
 	for (l = keyring->items; l != NULL; l = l->next) {
 		item = l->data;
-		gnome_keyring_proto_add_uint32 (buffer, item->id);
-		gnome_keyring_proto_add_uint32 (buffer, item->type);
+		gkr_buffer_add_uint32 (buffer, item->id);
+		gkr_buffer_add_uint32 (buffer, item->type);
 		
 		hashed = gnome_keyring_attributes_hash (item->attributes);
 
@@ -438,7 +440,7 @@ generate_file (GkrBuffer *buffer, GkrKeyring *keyring)
 		gkr_buffer_uninit (&to_encrypt);
 		return FALSE;
 	}
-	gnome_keyring_proto_add_uint32 (buffer, to_encrypt.len);
+	gkr_buffer_add_uint32 (buffer, to_encrypt.len);
 	gkr_buffer_append (buffer, to_encrypt.buf, to_encrypt.len);
 	gkr_buffer_uninit (&to_encrypt);
 	
@@ -457,11 +459,11 @@ decode_acl (GkrBuffer *buffer, gsize offset, gsize *offset_out, GList **out)
 	
 	acl = NULL;
 
-	if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset, &num_acs)) {
+	if (!gkr_buffer_get_uint32 (buffer, offset, &offset, &num_acs)) {
 		return FALSE;
 	}
 	for (i = 0; i < num_acs; i++) {
-		if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset, &x)) {
+		if (!gkr_buffer_get_uint32 (buffer, offset, &offset, &x)) {
 			goto bail;
 		}
 		if (!gnome_keyring_proto_get_utf8_string (buffer, offset, &offset,
@@ -481,7 +483,7 @@ decode_acl (GkrBuffer *buffer, gsize offset, gsize *offset_out, GList **out)
 			g_free (reserved);
 			goto bail;
 		}
-		if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset, &y)) {
+		if (!gkr_buffer_get_uint32 (buffer, offset, &offset, &y)) {
 			g_free (name);
 			g_free (path);
 			g_free (reserved);
@@ -515,7 +517,7 @@ remove_unavailable_item (gpointer key, gpointer dummy, GkrKeyring *keyring)
 	
 	g_assert (GKR_IS_KEYRING (keyring));
 	
-	item = gkr_keyring_find_item (keyring, id);
+	item = gkr_keyring_get_item (keyring, id);
 	if (item)
 		gkr_keyring_remove_item (keyring, item);
 }
@@ -577,13 +579,13 @@ update_keyring_from_data (GkrKeyring *keyring, GkrBuffer *buffer)
 	if (!gnome_keyring_proto_get_time (buffer, offset, &offset, &mtime)) {
 		goto bail;
 	}
-	if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset, &flags)) {
+	if (!gkr_buffer_get_uint32 (buffer, offset, &offset, &flags)) {
 		goto bail;
 	}
-	if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset, &lock_timeout)) {
+	if (!gkr_buffer_get_uint32 (buffer, offset, &offset, &lock_timeout)) {
 		goto bail;
 	}
-	if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset, &hash_iterations)) {
+	if (!gkr_buffer_get_uint32 (buffer, offset, &offset, &hash_iterations)) {
 		goto bail;
 	}
 	if (!gnome_keyring_proto_get_bytes (buffer, offset, &offset, salt, 8)) {
@@ -591,7 +593,7 @@ update_keyring_from_data (GkrKeyring *keyring, GkrBuffer *buffer)
 	}
 	
 	for (i = 0; i < 4; i++) {
-		if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset, &tmp)) {
+		if (!gkr_buffer_get_uint32 (buffer, offset, &offset, &tmp)) {
 			goto bail;
 		}
 		/* reserved bytes must be zero */
@@ -599,18 +601,18 @@ update_keyring_from_data (GkrKeyring *keyring, GkrBuffer *buffer)
 			goto bail;
 		}
 	}
-	if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset, &num_items)) {
+	if (!gkr_buffer_get_uint32 (buffer, offset, &offset, &num_items)) {
 		goto bail;
 	}
 
 	items = g_new0 (ItemInfo, num_items);
 
 	for (i = 0; i < num_items; i++) {
-		if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset,
+		if (!gkr_buffer_get_uint32 (buffer, offset, &offset,
 						     &items[i].id)) {
 			goto bail;
 		}
-		if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset,
+		if (!gkr_buffer_get_uint32 (buffer, offset, &offset,
 						     &items[i].type)) {
 			goto bail;
 		}
@@ -620,7 +622,7 @@ update_keyring_from_data (GkrKeyring *keyring, GkrBuffer *buffer)
 		}
 	}
 
-	if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset,
+	if (!gkr_buffer_get_uint32 (buffer, offset, &offset,
 					     &crypto_size)) {
 		goto bail;
 	}
@@ -671,7 +673,7 @@ update_keyring_from_data (GkrKeyring *keyring, GkrBuffer *buffer)
 				}
 				for (j = 0; j < 4; j++) {
 					guint32 tmp;
-					if (!gnome_keyring_proto_get_uint32 (buffer, offset, &offset, &tmp)) {
+					if (!gkr_buffer_get_uint32 (buffer, offset, &offset, &tmp)) {
 						goto bail;
 					}
 					/* reserved bytes must be zero */
@@ -717,7 +719,7 @@ update_keyring_from_data (GkrKeyring *keyring, GkrBuffer *buffer)
 		/* We've seen this id */
 		g_hash_table_remove (checks, GUINT_TO_POINTER (items[i].id));
 		
-		item = gkr_keyring_find_item (keyring, items[i].id);
+		item = gkr_keyring_get_item (keyring, items[i].id);
 		if (item == NULL) {
 			item = gkr_keyring_item_new (keyring, items[i].id, items[i].type);
 			gkr_keyring_add_item (keyring, item);
@@ -798,40 +800,43 @@ write_all (int fd, const guchar *buf, size_t len)
 	return 0;
 }
 
-static char*
-get_default_keyring_file_for_name (const char *keyring_name)
+static GQuark
+get_default_location_for_name (GQuark base_loc, const char *keyring_name)
 {
-	char *base;
-	char *filename;
+	gchar *path = NULL;
+	gchar *base, *filename;
 	int version;
-	char *path;
-	char *dir;
+	GQuark loc;
 
 	base = g_filename_from_utf8 (keyring_name, -1, NULL, NULL, NULL);
-	if (base == NULL) {
+	if (base == NULL)
 		base = g_strdup ("keyring");
-	}
 
-	dir = gkr_keyrings_get_dir ();
-	
 	version = 0;
 	do {
-		if (version == 0) {
-			filename = g_strdup_printf ("%s.keyring", base);
-		} else {
-			filename = g_strdup_printf ("%s%d.keyring", base, version);
-		}
+		g_free (path);
 		
-		path = g_build_filename (dir, filename, NULL);
+		if (version == 0) 
+			filename = g_strdup_printf ("%s/keyrings/%s.keyring", 
+			                            g_quark_to_string (base_loc), base);
+		else
+			filename = g_strdup_printf ("%s/keyrings/%s%d.keyring", 
+			                            g_quark_to_string (base_loc), base, version);
+
+		loc = gkr_location_from_string (filename);
 		g_free (filename);
+		
+		path = gkr_location_to_path (loc);
+		g_return_val_if_fail (path, 0);
 
 		version++;
 	} while (g_file_test (path, G_FILE_TEST_EXISTS));
 
 	g_free (base);
-	g_free (dir);
 	
-	return path;
+	loc = gkr_location_from_path (path);
+	g_free (path);
+	return loc;
 }
 
 /* -----------------------------------------------------------------------------
@@ -861,6 +866,9 @@ gkr_keyring_get_property (GObject *obj, guint prop_id, GValue *value,
 	case PROP_NAME:
 		g_value_set_string (value, keyring->keyring_name);
 		break;
+	case PROP_LOCATION:
+		g_value_set_uint (value, keyring->location);
+		break;
 	}
 }
 
@@ -892,7 +900,6 @@ gkr_keyring_finalize (GObject *obj)
 	GkrKeyring *keyring = GKR_KEYRING (obj);
 
 	g_free (keyring->keyring_name);
-	g_free (keyring->file);
 	g_assert (keyring->password == NULL);
 	
 	G_OBJECT_CLASS (gkr_keyring_parent_class)->finalize (obj);
@@ -912,6 +919,10 @@ gkr_keyring_class_init (GkrKeyringClass *klass)
 	g_object_class_install_property (gobject_class, PROP_NAME,
 		g_param_spec_string ("name", "Name", "Keyring Name",
 		                     NULL, G_PARAM_READABLE));
+		                     
+	g_object_class_install_property (gobject_class, PROP_LOCATION,
+		g_param_spec_uint ("location", "Location", "File Location",
+		                   0, G_MAXUINT, 0, G_PARAM_READABLE));
 	
 	signals[ITEM_ADDED] = g_signal_new ("item-added", GKR_TYPE_KEYRING, 
 			G_SIGNAL_RUN_FIRST, G_STRUCT_OFFSET (GkrKeyringClass, item_added),
@@ -929,7 +940,7 @@ gkr_keyring_class_init (GkrKeyringClass *klass)
  */
 
 GkrKeyring*
-gkr_keyring_new (const char *name, const char *path)
+gkr_keyring_new (const char *name, GQuark location)
 {
 	GkrKeyring *keyring;
 	
@@ -938,19 +949,19 @@ gkr_keyring_new (const char *name, const char *path)
 	keyring = g_object_new (GKR_TYPE_KEYRING, NULL);
 	
 	keyring->keyring_name = g_strdup (name);
-	keyring->file = g_strdup (path);
+	keyring->location = location;
 
 	return keyring;
 }
 
 GkrKeyring*
-gkr_keyring_create (const gchar *keyring_name, const gchar *password)
+gkr_keyring_create (GQuark base_loc, const gchar *keyring_name, const gchar *password)
 {
 	GkrKeyring *keyring;
 	
-	keyring = gkr_keyring_new (keyring_name, NULL);
+	keyring = gkr_keyring_new (keyring_name, 0);
 	if (keyring != NULL) {
-		keyring->file = get_default_keyring_file_for_name (keyring_name);
+		keyring->location = get_default_location_for_name (base_loc, keyring_name);
 		keyring->locked = FALSE;
 		keyring->password = gnome_keyring_memory_strdup (password);
 		gkr_keyring_save_to_disk (keyring);
@@ -982,7 +993,7 @@ gkr_keyring_get_new_id (GkrKeyring *keyring)
 }
 
 GkrKeyringItem*
-gkr_keyring_find_item (GkrKeyring *keyring, guint id)
+gkr_keyring_get_item (GkrKeyring *keyring, guint id)
 {
 	GkrKeyringItem *item;
 	GList *l;
@@ -996,6 +1007,22 @@ gkr_keyring_find_item (GkrKeyring *keyring, guint id)
 	return NULL;
 }
 
+GkrKeyringItem*  
+gkr_keyring_find_item (GkrKeyring *keyring, GnomeKeyringItemType type, 
+                       GnomeKeyringAttributeList *attrs)
+{    
+	GkrKeyringItem *item;
+	GList *l;
+	
+	for (l = keyring->items; l; l = g_list_next (l)) {
+		item = GKR_KEYRING_ITEM (l->data);
+		if (gkr_keyring_item_match (item, type, attrs, TRUE))
+			return item;
+	}
+	
+	return NULL;
+}
+
 void
 gkr_keyring_add_item (GkrKeyring* keyring, GkrKeyringItem* item)
 {
@@ -1039,39 +1066,58 @@ gkr_keyring_update_from_disk (GkrKeyring *keyring, gboolean force_reload)
 	char *contents = NULL;
 	gsize len;
 	gboolean success = FALSE;
+	gchar *file = NULL;
 
-	if (keyring->file == NULL)
-		return TRUE;
+	if (!keyring->location) {
+		success = TRUE;
+		goto done;
+	}
 
-	if (stat (keyring->file, &statbuf) < 0)
+	file = gkr_location_to_path (keyring->location);
+	if (!file)
+		goto done;
+		
+	if (stat (file, &statbuf) < 0) {
+		g_free (file);
 		return FALSE;
+	}
 
-	if (!force_reload && statbuf.st_mtime == keyring->file_mtime)
-		return TRUE;
+	if (!force_reload && statbuf.st_mtime == keyring->file_mtime) {
+		success = TRUE;
+		goto done;
+	}
 	keyring->file_mtime = statbuf.st_mtime;
 
-	if (!g_file_get_contents (keyring->file, &contents, &len, NULL))
-		return FALSE;
-
-	gkr_buffer_init_static (&buffer, (guchar*)contents, len);
+	if (!g_file_get_contents (file, &contents, &len, NULL))
+		goto done;
 	
+	gkr_buffer_init_static (&buffer, (guchar*)contents, len);
 	success = update_keyring_from_data (keyring, &buffer);
 	gkr_buffer_uninit (&buffer);
-	g_free (contents);
 
+done:
+	g_free (contents);
+	g_free (file);
 	return success;
 }
 
 gboolean 
 gkr_keyring_remove_from_disk (GkrKeyring *keyring)
 {
+	gchar *file;
 	int res;
 
 	/* Cannot remove session or memory based keyring */
-	if (keyring->file == NULL)
+	if (!keyring->location)
 		return FALSE;
-
-	res = unlink (keyring->file);
+		
+	file = gkr_location_to_path (keyring->location);
+	if (!file)
+		return FALSE;
+		
+	res = unlink (file);
+	g_free (file);
+	
 	return (res == 0);
 }
 
@@ -1084,21 +1130,26 @@ gkr_keyring_save_to_disk (GkrKeyring *keyring)
 	char *dirname;
 	char *template;
 	gboolean ret = TRUE;
+	gchar *file = NULL;
 	
-	if (keyring->locked) {
-		/* Can't save locked keyrings */
+	/* Can't save locked keyrings */
+	if (keyring->locked)
 		return FALSE;
-	}
 
-	if (keyring->file == NULL) {
-		/* Not file backed */
+	/* Not file backed */
+	if (!keyring->location)
 		return TRUE;
-	}
+		
+	file = gkr_location_to_path (keyring->location);
+	if (!file)
+		return FALSE;
 	
 	gkr_buffer_init_full (&out, 4096, g_realloc);
 
 	if (generate_file (&out, keyring)) {
-		dirname = g_path_get_dirname (keyring->file);
+		dirname = g_path_get_dirname (file);
+		if (g_mkdir_with_parents (dirname, S_IRWXU) < 0)
+			g_warning ("unable to create keyring dir");
 		template = g_build_filename (dirname, ".keyringXXXXXX", NULL);
 		
 		fd = g_mkstemp (template);
@@ -1109,10 +1160,9 @@ gkr_keyring_save_to_disk (GkrKeyring *keyring)
 			fsync (fd);
 #endif
 				close (fd);
-				if (rename (template, keyring->file) == 0) {
-					if (stat (keyring->file, &statbuf) == 0) {
+				if (rename (template, file) == 0) {
+					if (stat (file, &statbuf) == 0)
 						keyring->file_mtime = statbuf.st_mtime;
-					}
 				} else {
 					unlink (template);
 				}
@@ -1132,6 +1182,51 @@ gkr_keyring_save_to_disk (GkrKeyring *keyring)
 	}
 	
 	gkr_buffer_uninit (&out);
+	g_free (file);
 	return ret;
 }
 
+gboolean
+gkr_keyring_lock (GkrKeyring *keyring)
+{
+	if (keyring->locked)
+		return TRUE;
+
+	/* Never lock the session keyring */
+	if (!keyring->location)
+		return TRUE;
+
+	g_assert (keyring->password != NULL);
+	
+	gnome_keyring_free_password (keyring->password);
+	keyring->password = NULL;
+	if (!gkr_keyring_update_from_disk (keyring, TRUE)) {
+		/* Failed to re-read, remove the keyring */
+		g_warning ("Couldn't re-read keyring %s\n", keyring->keyring_name);
+		gkr_keyrings_remove (keyring);
+	}
+	
+	return TRUE;
+}
+
+gboolean
+gkr_keyring_unlock (GkrKeyring *keyring, const gchar *password)
+{
+	if (!keyring->locked)
+		return TRUE;
+		
+	g_assert (keyring->password == NULL);
+		
+	keyring->password = gnome_keyring_memory_strdup (password);
+	if (!gkr_keyring_update_from_disk (keyring, TRUE)) {
+		gnome_keyring_free_password (keyring->password);
+		keyring->password = NULL;
+	}
+	if (keyring->locked) {
+		g_assert (keyring->password == NULL);
+		return FALSE;
+	} else {
+		g_assert (keyring->password != NULL);
+		return TRUE;
+	}
+}
